#! /usr/bin/env python

# Moritz Helias
# adapted by Sacha van Albada Apr 2015

# Random recurrent network to measure cross correlation
# functions between exc and inh neurons.
# The incoming connectivity to each neuron is the same (in particular for exc.
# and inh. neurons).
# A pseudo-population of neurons receives stochastically the same input
# as excitatory and inhibitory neurons.
# The following cross-correlation functions are recorded:
# exc-exc, exc-inh, inh-inh, exc-pseudo, inh-pseudo

import nest

import time
import numpy
import scipy
import scipy.io
from numpy import exp, sqrt
from parameters_LIF_network_rate_comparison import *


def create_and_connect_correlometer(pop1, pop2):
    cd = nest.Create('correlation_detector')
    # start recording at tmax to avoid fall-off to the sides
    nest.SetStatus(cd, {'Tstart': tmax, 'delta_tau': tbin, 'tau_max': tmax})
    nest.SetDefaults('static_synapse', {'weight': 1.0, 'receptor_type': 0})
    nest.ConvergentConnect(pop1, cd, model='static_synapse')
    nest.SetDefaults('static_synapse', {'receptor_type': 1})
    nest.ConvergentConnect(pop2, cd, model='static_synapse')
    return cd


def write_correlation_data(cd, fn):
   
   if nest.GetStatus(cd, 'local')[0]:

      hs = list(nest.GetStatus(cd, 'histogram')[0])
      n1 = nest.GetStatus(cd, 'n_events')[0][0]
      n2 = nest.GetStatus(cd, 'n_events')[0][1]

      cdarr = numpy.array([n1, n2] + hs)
      numpy.savetxt(fn, cdarr)


def write_autocorrelation_data(cds, fn):

    hs = numpy.zeros(2*tmax/tbin+1)
    n1 = n2 = 0
    
    for cd in cds:
        if nest.GetStatus(cd, 'local')[0]: 
            hs += nest.GetStatus(cd, 'histogram')[0]
            n1 += nest.GetStatus(cd, 'n_events')[0][0]
            n2 += nest.GetStatus(cd, 'n_events')[0][1]

    cdarr = numpy.array([n1, n2] + list(hs))
    numpy.savetxt(fn + str(nest.Rank()), cdarr)


def simulate(scale_type, K_scaling, N_scaling, g, gamma, epsilon, J, NE_full, delay, nu, mu_ext_full, sigma_ext_full, N_rec, procs):

    print 'scale_type: ', scale_type

    NI_full = int(round(NE_full*gamma))
    CE_full = int(round(epsilon*NE_full))
    CI_full = int(round(epsilon*NI_full))

    if scale_type == 'none':
        CE = CE_full
        CI = CI_full
        NE = NE_full
        NI = NI_full
    else:
        CE = int(round(K_scaling*CE_full))
        CI = int(round(K_scaling*CI_full))
        NE = int(round(N_scaling*NE_full))
        NI = int(round(N_scaling*NI_full))

    # scale synaptic current so that same amount of charge
    # as in iaf_psc_delta with synaptic amplitude J in mV
    # Q = J/Cm here Cm = 1
    # I_max = Q/tauSyn

    if scale_type != 'none':
        mu_local_full = tauMem/1000. * nu * (CE_full * J - CI_full * g * J)
        sigma2_local_full = tauMem/1000. * nu * (CE_full * J**2 + CI_full * g**2 * J**2)

        mu_full = mu_local_full + mu_ext_full
        sigma2_full = sigma2_local_full + sigma_ext_full**2

    if scale_type == 'none':
        J_ex = J
        J_in = -g*J
    if scale_type == 'vreeswijk':
        J_ex = J/sqrt(K_scaling)
        J_in = -g*J/sqrt(K_scaling)
    if scale_type == 'helias':
        J_ex = J/K_scaling
        J_in = -g*J/K_scaling

    # determine rate of balanced Poisson drive to generate the desired variance
    # and the required DC drive to recover the mean input
    if scale_type == 'none':
        nu_ext = sigma_ext_full**2/(2 * tauMem/1000. * J**2)
        mu = mu_ext_full
    else:
        mu_local = tauMem/1000. * nu * (CE * J_ex + CI * J_in)
        sigma2_local = tauMem/1000. * nu * (CE * J_ex**2 + CI * J_in**2)
     
        nu_ext = (sigma2_full - sigma2_local)/(2 * tauMem/1000. * J_ext**2)
        mu = mu_full - mu_local
    
    print 'external Poisson rate: ', nu_ext
    print 'DC input: ', mu

    nest.ResetKernel()

    nest.SetKernelStatus({'resolution': dt, 'total_num_virtual_procs': procs, 'overwrite_files': True, 'print_time': False})
    nest.sr("/RandomConvergentConnect << /allow_multapses false >> SetOptions")


    print "Building network"

    neuron_params= {"C_m":        1.0,
                    "tau_m":      tauMem,
                    "t_ref":      tauRef,
                    "tau_syn_ex": tauSyn,
                    "tau_syn_in": tauSyn,
                    "E_L":        V_r,
                    "V_reset":    V_r,
                    "V_th":       theta,
                    "I_e": 	  mu/tauMem}

    nest.SetDefaults(nrn_model, neuron_params)

    nodes_ex=nest.Create(nrn_model, NE)
    nodes_in=nest.Create(nrn_model, NI)
    nodes_pseudo=nest.Create(nrn_model, 2*N_rec)

    nest.CopyModel("static_synapse", "excitatory", {"weight":J_ex/tauSyn, "delay":delay})
    nest.CopyModel("static_synapse", "inhibitory", {"weight":J_in/tauSyn, "delay":delay})

    # external poisson input

    nest.SetDefaults("poisson_generator_ps", {"rate": nu_ext})
    noise_E = nest.Create("poisson_generator_ps")
    noise_I = nest.Create("poisson_generator_ps")

    print "Connecting devices."

    nest.DivergentConnect(noise_E, nodes_ex, J_ext/tauSyn, delay)
    nest.DivergentConnect(noise_E, nodes_in, J_ext/tauSyn, delay)
    nest.DivergentConnect(noise_E, nodes_pseudo, J_ext/tauSyn, delay)

    nest.DivergentConnect(noise_I, nodes_ex, -J_ext/tauSyn, delay)
    nest.DivergentConnect(noise_I, nodes_in, -J_ext/tauSyn, delay)
    nest.DivergentConnect(noise_I, nodes_pseudo, -J_ext/tauSyn, delay)

    if N_rec_sd > 0:
        espikes=nest.Create("spike_detector")
        ispikes=nest.Create("spike_detector")

        nest.SetStatus([espikes],[{"label": datapath + "brunel-py-ex",
                                   "withtime": True,
                                   "withgid": True,
                                   "to_memory": False,
                                   "to_file": True}])

        nest.SetStatus([ispikes],[{"label": datapath + "brunel-py-in",
                                   "withtime": True,
                                   "withgid": True,
                                   "to_memory": False,
                                   "to_file": True}])

        nest.ConvergentConnect(range(1,N_rec_sd+1), espikes, model="excitatory")
        nest.ConvergentConnect(range(NE+1,NE+1+N_rec_sd), ispikes, model="excitatory")

    print "Connecting network."

    # We now iterate over all neuron IDs, and connect the neuron to
    # the sources from our array. The first loop connects the excitatory neurons
    # and the second loop the inhibitory neurons.

    print "Excitatory connections"

    nest.RandomConvergentConnect(nodes_ex, nodes_ex+nodes_in+nodes_pseudo, CE, model="excitatory")

    print "Inhibitory connections"

    nest.RandomConvergentConnect(nodes_in, nodes_ex+nodes_in+nodes_pseudo, CI, model="inhibitory")


    ############################################################
    ## connect recording neurons and correlation detectors
    ###########################################################

    # auto correlations

    cd_acf_ee = []
    cd_acf_ii = []

    for n in xrange(N_rec_acf):
        cd_acf_ee.append( create_and_connect_correlometer([nodes_ex[n]], [nodes_ex[n]]) )
        cd_acf_ii.append( create_and_connect_correlometer([nodes_in[n]], [nodes_in[n]]) )

    # cross correlations
    cd_ee = create_and_connect_correlometer(nodes_ex[:N_rec], nodes_ex[N_rec:2*N_rec])
    cd_ii = create_and_connect_correlometer(nodes_in[:N_rec], nodes_in[N_rec:2*N_rec])
    cd_ei = create_and_connect_correlometer(nodes_ex[:N_rec], nodes_in[:N_rec])
    cd_ep = create_and_connect_correlometer(nodes_ex[:N_rec], nodes_pseudo[:N_rec])
    cd_ip = create_and_connect_correlometer(nodes_in[:N_rec], nodes_pseudo[:N_rec])
    cd_pp = create_and_connect_correlometer(nodes_pseudo[:N_rec], nodes_pseudo[N_rec:2*N_rec])


    print "Simulating."

    nest.Simulate(simtime)

    suffix = scale_type + 'K_scaling' + str(K_scaling) + 'NE' + str(NE) + 'J' + str(J) + 'delay' + str(delay) + 'mu_ext_full' + str(mu_ext_full) + 'sigma_ext_full' + str(sigma_ext_full) + '.dat'
    print 'suffix: ', suffix
    write_correlation_data(cd_ee, datapath + 'corr_ee' + suffix)
    write_correlation_data(cd_ii, datapath + 'corr_ii' + suffix)
    write_correlation_data(cd_ei, datapath + 'corr_ei' + suffix)
    write_correlation_data(cd_ep, datapath + 'corr_ep' + suffix)
    write_correlation_data(cd_ip, datapath + 'corr_ip' + suffix)
    write_correlation_data(cd_pp, datapath + 'corr_pp' + suffix)

    write_autocorrelation_data(cd_acf_ee, datapath + 'acf_ee' + suffix)
    write_autocorrelation_data(cd_acf_ii, datapath + 'acf_ii' + suffix)

