# Sacha van Albada, Feb 2015

# Covariance functions for different ways of downscaling 
# the microcircuit model of Potjans & Diesmann (Cereb Cortex, 2014) 

from plotfuncs import *
import csv
import numpy

base_dir = '../data/spiking/microcircuit/'

figurepath = '../figures/'

mycols = [black, '#1d771d', '#a8d57b']

pops = ['L23E', 'L23I', 'L4E', 'L4I', 'L5E', 'L5I', 'L6E', 'L6I']

# parameters of covariance functions
tau_max = 100.
delta_tau = 0.5
cov_t = numpy.arange(-tau_max,tau_max+delta_tau/2,delta_tau)
zero_index = numpy.where(cov_t==0)[0][0]

# in-degree scaling factor for downscaled models
K_scaling = 0.9

scale_types = ['w_and_ext', 'w_mean_var_ext']

rates_w_and_ext_full = {
'L23I': 3.85465381782,
'L23E': 1.05332982223,
'L4I': 6.82597724674,
'L4E': 2.31199229336,
'L5I': 8.68397105343,
'L5E': 5.33064889784,
'L6I': 8.59130469565,
'L6E': 1.75332748891}

# rates for N_scaling = 1, K_scaling=0.9:
rates_w_and_ext_downscaled = {
'L23I': 5.06131646228,
'L23E': 1.16732944224,
'L4I': 7.91664027787,
'L4E': 3.08732304226,
'L5I': 11.4426285246,
'L5E': 8.36997210009,
'L6I': 11.215295949,
'L6E': 2.02999323336}

# rates for N_scaling = 1, K_scaling=0.9:
rates_w_mean_var_ext_downscaled = {
'L23I': 4.47065176449,
'L23E': 1.08466305112,
'L4I': 6.9813100623,
'L4E': 2.71465761781,
'L5I': 10.4032986557,
'L5E': 8.07130642898,
'L6I': 10.1519661601,
'L6E': 1.93932686891}

# load covariances

all_pop_cov = {}
for scale_type in scale_types:
    all_pop_cov[scale_type] = {}
    for pop1 in pops:
        all_pop_cov[scale_type][pop1] = {}
        for pop2 in pops:
            all_pop_cov[scale_type][pop1][pop2] = numpy.zeros([2, len(cov_t)]) + numpy.nan

for scale_type in scale_types:
    for i, K in enumerate([K_scaling, 1.0]):
        with open(base_dir + scale_type + '/' + '1.0_' + str(K) + '/covariances.dat', 'rb') as csvfile:
            l = csv.reader(csvfile, delimiter='\n')
            for row in l:
                if row:
                    if row[0][0]=='L':
                        pop1 = row[0].split('-')[0].replace(' ','')
                        pop2 = row[0].split('-')[1].replace(' ','')
                        index = 0
                    if row[0][0] not in ('t', 'd', 'L'):
                        all_pop_cov[scale_type][pop1][pop2][i][index] = float(row[0])
                        index += 1

pop1 = 'L23E'
pop2 = 'L23E'

# covariances were based on counts from all neurons
num_neurons = {'L23E':20683, 'L23I':5834,
  'L4E':21915, 'L4I':5479,
  'L5E':4850,  'L5I':1065,
  'L6E':14395, 'L6I':2948}

# time over which events were counted
T = 30000.-0.1-tau_max


# normalize to get c^2/r^2
# if you want to plot covariances for scaled N, adjust the normalization below
norm = (num_neurons[pop1]*num_neurons[pop2]*delta_tau/1000.*T/1000.)

cov_full = all_pop_cov['w_and_ext'][pop1][pop2][1]/norm/rates_w_and_ext_full[pop1]/rates_w_and_ext_full[pop2]
cov_w_and_ext = all_pop_cov['w_and_ext'][pop1][pop2][0]/norm/rates_w_and_ext_downscaled[pop1]/rates_w_and_ext_downscaled[pop2]
cov_w_mean_var_ext = all_pop_cov['w_mean_var_ext'][pop1][pop2][0]/norm/rates_w_mean_var_ext_downscaled[pop1]/rates_w_mean_var_ext_downscaled[pop2]

scale = 1.0
width = 6.83 
n_horz_panels = 2
n_vert_panels = 1

panel_factory = create_fig_plos(1, scale, width, n_horz_panels, n_vert_panels)

xmin = -50
xmax = 50
ymin = -0.06
ymax = 0.09
ytcks = [-0.04, 0.0, 0.04, 0.08]

axA = panel_factory.new_panel(0, 0, 'A')
plot(cov_t, cov_full, color=mycols[0], label='full scale')
plot(cov_t, cov_w_and_ext, color=mycols[1], label=r'$K=0.9 K_0$')
f0 = cov_full[zero_index]/cov_w_and_ext[zero_index]
plot(cov_t, cov_w_and_ext*f0, color=mycols[2], label='fit to c(0)')

rcParams['legend.fontsize'] = scale*8

xlim([xmin,xmax])
ylim([ymin,ymax])
yticks(ytcks)
xlabel(r'$\Delta$' + ' (ms)')
ylabel(r'$\frac{c}{r^2}$')
title(r'$J \propto \frac{1}{\sqrt{K}}$')
legend(loc=1)

axB = panel_factory.new_panel(1, 0, 'B')
plot(cov_t, cov_full, color=mycols[0], label='full scale')
plot(cov_t, cov_w_mean_var_ext, color=mycols[1], label=r'$K=0.9 K_0$')
f0 = cov_full[zero_index]/cov_w_mean_var_ext[zero_index]
plot(cov_t, cov_w_mean_var_ext*f0, color=mycols[2], label='fit to c(0)')

xlim([xmin,xmax])
ylim([ymin,ymax])
yticks(ytcks)
xlabel(r'$\Delta$' + ' (ms)')
ylabel(r'$\frac{c}{r^2}$')
title(r'$J \propto \frac{1}{K}$')

savefig(figurepath + 'fig4.eps')

