


import numpy as np
import numpy.linalg as la
import scipy.special as sps
from scipy.optimize import fsolve
from scipy.integrate import quad


# Moritz Helias 2013-05-31

# index convention
E = 0
I = 1
X = 2

EE = 0
EI = 1
II = 2
EX = 3
IX = 4


class ginzburg:


    def __init__(self, N, J, p, m_x, theta):
        self.N = N
        self.J = J
        self.p = p
        self.K = np.vstack([N * p, N * p]) # for each population one row
        self.m_x = m_x
        self.theta = theta
        

    def phi(self, mu, sigma):
        return 0.5*(1-sps.erf(-(mu-self.theta)/(np.sqrt(2.)*sigma)))

    
    def p_mu(self, m, mu, sigma, q, x):
        """probability to have mean input x"""
        q1 = np.hstack((q, self.m_x*self.m_x))
        m1 = np.hstack((m, self.m_x))
        KJ2 = self.K*self.J*self.J        
        delta_mu2 = np.dot( KJ2, (q1 - self.p*m1**2) )
        p_x = 1./np.sqrt(2.*np.pi*delta_mu2) * np.exp( -(mu - x)**2 / (2.*delta_mu2) )

        phi_prime_m1 = np.sqrt(2.*np.pi)*sigma * np.exp( (x - self.theta)**2 / (2.*sigma**2) )
        return p_x * phi_prime_m1 


    def m2(self, m, mu, sigma):
        """calculate the second moment q = [m^2] of the activity distribution"""
        
        # need to calculate self-consistently
        m1 = np.hstack( (m, self.m_x*self.m_x) ) 

        q0 = m1*m1

        print 'mu = ', mu
        delta_mu2 = np.dot( self.J*self.J, self.p*(1.-self.p) * self.N*q0 + (self.N*self.p)**2 * (q0 - m1**2) )        
        print 'delta_mu2 =', delta_mu2

        def f(x):
            return 1./np.sqrt(2.*np.pi*delta_mu2) * np.exp( -(mu - x)**2 / (2.*delta_mu2) ) * self.phi(x, sigma)**2

        def g(x):
            return 1./np.sqrt(2.*np.pi*delta_mu2) * np.exp( -(mu - x)**2 / (2.*delta_mu2) ) * self.phi(x, sigma)

        # integrate for each population separately
        q = np.array([quad(lambda x : f(x)[i], mu[i] - 6.*delta_mu2[i], mu[i] + 6.*delta_mu2[i])[0] for i in range(len(mu)) ])
        r = np.array([quad(lambda x : g(x)[i], mu[i] - 6.*delta_mu2[i], mu[i] + 6.*delta_mu2[i])[0] for i in range(len(mu)) ])

        print 'q =', q
        print 'r =', r 
        return q

    def m2_mu_sigma(self, mu, sigma, delta_mu):
        """calculate the second moment q = [m^2] of the activity distribution"""
        
        # need to calculate self-consistently

        def f(x):
            return 1./(np.sqrt(2.*np.pi)*delta_mu) * np.exp( -(mu - x)**2 / (2.*delta_mu**2) ) * self.phi(x, sigma)**2

        def g(x):
            return 1./(np.sqrt(2.*np.pi)*delta_mu) * np.exp( -(mu - x)**2 / (2.*delta_mu**2) ) * self.phi(x, sigma)

        # integrate for each population separately
        q = np.array([quad(lambda x : f(x)[i], mu[i] - 6.*delta_mu[i], mu[i] + 6.*delta_mu[i])[0] for i in range(len(mu)) ])
        r = np.array([quad(lambda x : g(x)[i], mu[i] - 6.*delta_mu[i], mu[i] + 6.*delta_mu[i])[0] for i in range(len(mu)) ])

        print 'q =', np.sqrt(q - r*r)
        print 'r =', r 
        return r, q
        

    def mu_sigma_from_m_q(self, m, q):
        """calculate the mean and variance of the input to a neuron
        given the mean activity and its second moment q"""
        
        m = np.hstack((m, self.m_x))
        q = np.hstack((q, self.m_x*self.m_x))

        KJ = self.K*self.J
        KJ2 = self.K*self.J*self.J

        # vector of variances
        # here we assume that the indegree is binomially distributed

        # This expression should be the same as in van vreeswijk 1998
        delta_mu2 = np.dot( KJ2, (q - self.p*m**2) )
        print 'delta_mu2 =', delta_mu2
        
        a = m - q


        mu = np.dot(KJ,m)

        # take into account correlations in the input to determine variance
        sigma2 = np.dot(KJ2, a)

        return mu, np.sqrt(np.abs(sigma2)), np.sqrt(np.abs(delta_mu2))


    def mu_sigma_from_m_q_corr(self, m, q, c):
        """calculate the mean and variance of the input to a neuron
        given the mean activity and its second moment q"""
        
        m = np.hstack((m, self.m_x))
        q = np.hstack((q, self.m_x*self.m_x))

        C = np.array([ [ c[EE], c[EI], c[EX] ],
                       [ c[EI], c[II], c[IX] ],
                       [ c[EX], c[IX], 0.    ] ])


        KJ = self.K*self.J
        KJ2 = self.K*self.J*self.J

        # vector of variances
        # here we assume that the indegree is binomially distributed

        delta_mu2 = np.dot( KJ2, (q - self.p*m**2) )
        print 'delta_mu2 =', delta_mu2
        
        a = m - q


        mu = np.dot(KJ,m)

        # take into account correlations in the input to determine variance
        sigma2_c = np.diag( np.dot( np.dot(KJ, C), KJ.T ))
        sigma2_a = np.dot(KJ2, a)
        sigma2 = sigma2_a + sigma2_c

        return mu, np.sqrt(np.abs(sigma2)), np.sqrt(np.abs(delta_mu2))


    def solve_self_consistently_m2(self, m0, q0):
        """determine self-consistent rates and rate variability
        neglecting effect of correlations on input statistics into account"""

        
        def m_zero_q(x):
            print 'iteration', x
            m = x[0:2]
            q = x[2:]
            
            mu, sigma, delta_mu = self.mu_sigma_from_m_q(m, q)

            def f(x):
                return 1./(np.sqrt(2.*np.pi)*delta_mu) * np.exp( -(mu - x)**2 / (2.*delta_mu**2) ) * self.phi(x, sigma)**2

            # integrate for each population separately
            new_q = np.array( [quad(lambda x : f(x)[i], mu[i] - 6.*delta_mu[i], mu[i] + 6.*delta_mu[i])[0] for i in range(len(mu)) ] )

            new_m = self.phi ( mu, np.sqrt( sigma**2 + delta_mu**2 ) )

            print 'new_m=', new_m
            print new_q
            out = np.hstack( (new_m - m, new_q - q) )            
            print out
            return out

        mq0 = fsolve(m_zero_q, np.hstack( (m0, q0) ))
        

        return mq0[:2], mq0[2:]        



    def mu_sigma_from_m_corr(self, m, c):
        """calculate the mean and variance of the input to a neuron
        given the mean activity and covariances"""
        
        m = np.hstack((m, self.m_x))
        C = np.array([ [ c[EE], c[EI], c[EX] ],
                       [ c[EI], c[II], c[IX] ],
                       [ c[EX], c[IX], 0.    ] ])

        # vector of variances
        # here we assume that the indegree is binomially distributed
        a = m*(1. - m)

        KJ = self.K*self.J
        KJ2 = self.K*self.J*self.J

        mu = np.dot(KJ,m)

        # take into account correlations in the input to determine variance
        sigma2_m = np.dot(KJ2, a)
        sigma2_c = np.diag( np.dot( np.dot(KJ, C), KJ.T ))
        
        sigma2 = np.sqrt(sigma2_m + sigma2_c)

        return mu, sigma2


    def susceptibility_corr(self, m, c):

        mu, sigma = self.mu_sigma_from_m_corr(m, c)

        S = 1./(np.sqrt(2.*np.pi)*sigma) * np.exp( -(mu-self.theta)**2 / (2.*sigma**2) )

        return S


    def susceptibility_m2_corr(self, m, q, c):
        
        mu, sigma, delta_mu = self.mu_sigma_from_m_q_corr(m, q, c)

        sigma2 = sigma**2 + delta_mu**2

        S = 1./np.sqrt(2.*np.pi*sigma2) * np.exp( -(mu-self.theta)**2 / (2.*sigma2) )

        return S


    def correlations_q0(self, m, S):
        """solve the linear system of equations determining the correlations
        given we know the susceptibility"""
        W = ((self.J*self.K).T * S).T

        N = self.N

        # we need to adapt a to include the
        # distribution of the mean activities of
        # the neurons
        a = m*(1.-m)
        c = np.zeros(5, dtype=float)

        pref = a[X]/self.N[X] / ( (2.-W[E,E])*(2.-W[I,I]) - W[E,I]*W[I,E] )    
        c[EX] = pref * ( (2. - W[I,I])*W[E,X] + W[E,I]*W[I,X] )
        c[IX] = pref * ( (2. - W[E,E])*W[I,X] + W[I,E]*W[E,X] )

        eps = 1.
        M = np.matrix([[1.*eps - W[E,E], - W[E,I], 0.],
                       [-0.5*W[I,E], 1.*eps - 0.5*(W[E,E]+W[I,I]), -0.5*W[E,I]],
                       [0., -W[I,E], 1.*eps - W[I,I]]])

        b = np.array([W[E,E]*a[E]/N[E], 0.5*(W[I,E]*a[E]/N[E] + W[E,I]*a[I]/N[I]), W[I,I]*a[I]/N[I]])

        b1 = np.array( [W[E,X]*c[EX], 0.5*(W[E,X]*c[IX] + W[I,X]*c[EX]), W[I,X]*c[IX]] )

        c[:EX] = np.linalg.solve(M, b + b1)

        return c


    def correlations(self, m, q, S, eps=1.):
        """solve the linear system of equations determining the correlations
        given we know the susceptibility"""
        W = ((self.J*self.K).T * S).T

        N = self.N

        # q includes the distribution of the mean activities of
        # the neurons
        a = m - q
        c = np.zeros(5, dtype=float)

        pref = a[X]/self.N[X] / ( (2.-W[E,E])*(2.-W[I,I]) - W[E,I]*W[I,E] )    
        c[EX] = pref * ( (2. - W[I,I])*W[E,X] + W[E,I]*W[I,X] )
        c[IX] = pref * ( (2. - W[E,E])*W[I,X] + W[I,E]*W[E,X] )

        M = np.matrix([[1.*eps - W[E,E], - W[E,I], 0.],
                       [-0.5*W[I,E], 1.*eps - 0.5*(W[E,E]+W[I,I]), -0.5*W[E,I]],
                       [0., -W[I,E], 1.*eps - W[I,I]]])

        b = np.array([W[E,E]*a[E]/N[E], 0.5*(W[I,E]*a[E]/N[E] + W[E,I]*a[I]/N[I]), W[I,I]*a[I]/N[I]])

        b1 = np.array( [W[E,X]*c[EX], 0.5*(W[E,X]*c[IX] + W[I,X]*c[EX]), W[I,X]*c[IX]] )

        c[:EX] = np.linalg.solve(M, b + b1)

        return c


    def correlations1(self, m, q, S, eps=1.):
        """solve the linear system of equations determining the correlations
        given we know the susceptibility. Split into term due to external drive and intrinsic fluctuations"""
        W = ((self.J*self.K).T * S).T

        N = self.N

        # q includes the distribution of the mean activities of
        # the neurons
        a = m - q
        c0 = np.zeros(5, dtype=float)
        c1 = np.zeros(5, dtype=float)
        
        pref = a[X]/self.N[X] / ( (2.-W[E,E])*(2.-W[I,I]) - W[E,I]*W[I,E] )    
        c0[EX] = pref * ( (2. - W[I,I])*W[E,X] + W[E,I]*W[I,X] )
        c0[IX] = pref * ( (2. - W[E,E])*W[I,X] + W[I,E]*W[E,X] )

        M = np.matrix([[1.*eps - W[E,E], - W[E,I], 0.],
                       [-0.5*W[I,E], 1.*eps - 0.5*(W[E,E]+W[I,I]), -0.5*W[E,I]],
                       [0., -W[I,E], 1.*eps - W[I,I]]])

        b = np.array([a[E]/N[E], 0., a[I]/N[I]])

        b1 = np.array( [W[E,X]*c0[EX], 0.5*(W[E,X]*c0[IX] + W[I,X]*c0[EX]), W[I,X]*c0[IX]] )

        c0[:EX] = np.linalg.solve(M, b1)
        c1[:EX] = np.linalg.solve(M, b)
        c0[EE] -= a[E]/N[E]
        c0[II] -= a[I]/N[I]

        return c0, c1


    def solve_self_consistently_corr(self, m0):
        """determine self-consistent rates and correlations
        taking effect of correlations on input statistics into account"""

        c = np.zeros(5)
        
        def m_zero_corr(m):

            mu, sigma = self.mu_sigma_from_m_corr(m, c)

            new_m = self.phi(mu, sigma)

            return new_m - m

        err = 1.
        while err > 1e-15:
            m0 = fsolve(m_zero_corr, m0)
            S = self.susceptibility_corr(m0, c)
            c_new = self.correlations_q0(np.hstack((m0, self.m_x)), S)
            err = la.norm(c_new - c)
            c = c_new

        print 'error ', err

        return m0, c_new


    def solve_self_consistently(self, m0):
        """determine self-consistent rates, neglect
        effect of correlations on input statistics"""

        c = np.zeros(5)
        
        def m_zero_corr(m):
            mu, sigma = self.mu_sigma_from_m_corr(m, c)

            new_m = 0.5*(1.-sps.erf(-(mu-self.theta)/(np.sqrt(2.)*sigma)))

            return new_m - m

        m0 = fsolve(m_zero_corr, m0)
        return m0



    def correlations_t(self, m, q, S, c, tau_m):
        
        j_ee = S[E]*(self.K*self.J)[E,E]
        j_ii = S[I]*(self.K*self.J)[I,I]
        j_ie = S[I]*(self.K*self.J)[I,E]
        j_ei = S[E]*(self.K*self.J)[E,I]

	sqr=np.sqrt(complex((j_ee-j_ii)**2+4*j_ei*j_ie,0))
	xp=0.5*(j_ee+j_ii+sqr)
	xm=0.5*(j_ee+j_ii-sqr)
	print 'xp, xm:', xp, xm, j_ee+j_ii
        
        ce0 = c[EE] + (m[E]-q[E]) / self.N[E]
        ci0 = c[II] + (m[I]-q[I]) / self.N[I]
        cie0 = c[EI]
        cae, cai = m - q 

	def c_ee(t):
		return ((-j_ii*ce0+j_ei*cie0)*(np.exp(-abs(t)*(1-xp)/tau_m)-np.exp(-abs(t)*(1-xm)/tau_m))+ce0 * (xp*np.exp(-abs(t)*(1-xp)/tau_m)-xm*np.exp(-abs(t)*(1-xm)/tau_m)))/(xp-xm)

	def c_ii(t):
		return ((-j_ee*ci0+j_ie*cie0)*(np.exp(-abs(t)*(1-xp)/tau_m)-np.exp(-abs(t)*(1-xm)/tau_m))+ci0 * (xp*np.exp(-abs(t)*(1-xp)/tau_m)-xm*np.exp(-abs(t)*(1-xm)/tau_m)))/(xp-xm)

	def c_ene(t):
		return -cae/self.N[E] * np.exp(-abs(t)/tau_m)+c_ee(t)

	def c_ini(t):
		return -cai/self.N[I] * np.exp(-abs(t)/tau_m)+c_ii(t)

	def c_ie(t):
		if t<0:    
			return ((-j_ii*cie0+j_ei*ci0)*(np.exp(-abs(t)*(1-xp)/tau_m)-np.exp(-abs(t)*(1-xm)/tau_m))+cie0 * (xp*np.exp(-abs(t)*(1-xp)/tau_m)-xm*np.exp(-abs(t)*(1-xm)/tau_m)))/(xp-xm)
		else:
			return ((-j_ee*cie0+j_ie*ce0)*(np.exp(-abs(t)*(1-xp)/tau_m)-np.exp(-abs(t)*(1-xm)/tau_m))+cie0 * (xp*np.exp(-abs(t)*(1-xp)/tau_m)-xm*np.exp(-abs(t)*(1-xm)/tau_m)))/(xp-xm)


	def c_ei(t):
		return c_ie(-t)

	return c_ene, c_ini, c_ei, c_ie

