# Moritz Helias, 2014

# Covariances and their pole structure for Hawkes or LIF networks

import numpy
import scipy
import scipy.integrate
import pylab
import mpmath as mpf

class hawkesnet:

    def __init__(self, NE, NI, epsilon, g, gamma, tau, delay, nu, poles_max):
        self.NE = NE
        self.NI = NI
        self.epsilon = epsilon
        self.gamma = gamma
        self.tau = tau
        self.d = delay
        self.g = g

        self.CE = epsilon*NE
        self.CI = epsilon*NI

        self.nu = nu # rate parameter of each neuron

        self.nu0 = nu / (1.0-gamma*(self.CE-g*self.CI))         # firing rate per neuron
        print 'nu0 (theo) =', self.nu0

        self.poles_max = poles_max


    
    def G_(self, s):
        return self.gamma * numpy.exp(-self.d*s)/(1.0 + s*self.tau)

 
    def get_poles_(self, K, branch_max):
        # Calculate poles of recurrent transfer function
        # using the Lambert W function. branch_max specifies the
        # positive maximum branch until which to go. The first two
        # entries are the solutions from the principal branch, which are
        # real for z > -1/e and which are conjugate pairs for z < -1/e.
        #
        #K = self.gamma*(self.CE - self.g*self.CI)

        z0s = []
        for branch in range(-1, branch_max):
            z0 = -1.0/self.tau + 1.0 / self.d * mpf.lambertw(K*self.d/self.tau*numpy.exp(self.d/self.tau), branch)

            # check: should be zero
            z0 = complex(z0)
            zr = 1.0 - K/self.gamma*self.G_(z0)
            assert(abs(zr) < 1e-10)
            # if assertion does not break, it works
            z0s.append(z0)

            # add missing conjugate complex poles for branch >= 1            
            if branch >= 1:
                zr = 1.0 - K/self.gamma*self.G_(z0.conjugate())
                assert(abs(zr) < 1e-10)
                z0s.append(z0.conjugate())

        return z0s

   
    def cross_cov_matrix2(self, gamma_eff, g_eff, nu0):
        """Calculates the cross covariance matrix for positive t
        according to equation eq:cross_cov_time without delta peak."""

        K = gamma_eff*(self.CE - g_eff*self.CI)        

        s0s = self.get_poles_(K, self.poles_max)
    
        eps = 1.0*self.CE/self.NE
        assert(abs ( eps - 1.0*self.CI/self.NI ) < 1e-8)

        m21 = gamma_eff * eps * numpy.matrix([ [1.0, -g_eff], [1.0, -g_eff] ])

        m3 = eps * gamma_eff**2 * (self.CE + g_eff**2*self.CI) * numpy.matrix( [ [1.0, 1.0], [1.0, 1.0] ] )
        
        def prefactor(s0):
            return nu0*1000.0 / (self.tau + self.d*(1.0+s0*self.tau) )

        def prefactor1(s0):
            return 1.0 / (1.0 - s0*self.tau - K*numpy.exp(s0*self.d))

        def Q(t_prime):
            # use symmetry: Q^T(t) = Q(-t) to calculate for negative t
            if t_prime < 0:
                t = -t_prime
            else:
                t = t_prime

            q = numpy.matrix([[0.0, 0.0], [0.0, 0.0]])

            # all poles are returned, conjugate pairs are in the list explicitly
            for s0 in s0s:
                q += prefactor(s0) * ( (t>=self.d) * m21 * numpy.exp(s0*(t-self.d))\
                                     + prefactor1(s0) * (t>=0) * m3 * numpy.exp(s0*t) )
                                        
            if t_prime < 0.0:
                return q.transpose()
            else:
                return q

        return Q

