# Sacha van Albada, Nov 2014

from network_params import *
from sim_params import *
from scaling import *
from connectivity import *
from pyNN.random import NumpyRNG, RandomDistribution
import numpy
import sys

class Network:

    def __init__(self, sim):
        return None

    def setup(self,sim) :

        # Create matrix of synaptic weights
        self.w = create_weight_matrix()
        model = getattr(sim, neuronmodel)
        if parallel_safe:
          script_rng = NumpyRNG(seed=seeds[0], parallel_safe=True)
        else:
          script_rng = NumpyRNG(seed=seeds[sim.rank()], parallel_safe=False)

        V_dist = sim.RandomDistribution('normal', [V0_mean, V0_sd], rng=script_rng)

        # create correlation recording device
        if record_corr and simulator == 'nest':
            sim.nest.SetDefaults('correlomatrix_detector',{'delta_tau' : 0.5*time_scaling})
            self.corr_detector = sim.nest.Create('correlomatrix_detector')
            sim.nest.SetStatus(self.corr_detector, {'N_channels': n_layers*n_pops_per_layer, 'tau_max': tau_max, 'Tstart': tau_max})

        # create cortical populations
        self.pops = {}
        for layer in layers:
            self.pops[layer] = {}
            for pop in pops:
                self.pops[layer][pop] = sim.Population(int(num_neurons[layer][pop]*N_scaling), model, cellparams=neuron_parameters)
                self.pops[layer][pop].initialize('v', V_dist)
                this_pop = self.pops[layer][pop]
                # spike recording
                if record_fraction:
                   num_spikes = round(this_pop.size*frac_record_spikes)
                else:
                   num_spikes = n_record
                this_pop[0:num_spikes].record(to_file=write_spikes)
                # membrane potential recording

                if record_v:
                    if record_fraction:
                        num_v = round(this_pop.size*frac_record_v)
                    else:
                        num_v = n_record_v
                    this_pop[0:num_v].record_v(to_file=write_v)
                # correlation recording
                if record_corr and simulator == 'nest':
                    index = structure[layer][pop]
                    if simulator == 'nest' :
                        sim.nest.SetDefaults("static_synapse", {'receptor_type': index})
                        #sim.nest.ConvergentConnect(list(this_pop.all_cells), self.corr_detector)
                        sim.nest.Connect(list(this_pop.all_cells), self.corr_detector, {'rule': 'all_to_all'})

        if simulator == 'nest' :
            # reset receptor_type
            sim.nest.SetDefaults("static_synapse", {'receptor_type': 0})
        # create thalamic population
        if thalamic_input:
            self.thalamic_population = sim.Population(thal_params['n_thal'], sim.SpikeSourcePoisson, {'rate': thal_params['rate'], 'start': thal_params['start'], 'duration': thal_params['duration']})
            w_thal_dist = RandomDistribution(distribution='normal', parameters=[w_mean, w_mean*w_rel], rng=script_rng, boundaries=(0,1000*w_mean), constrain='redraw')



        # compute DC input before scaling
        if 'DC' in input_types:
            if DC_automatic:
                self.DC_amp = {}
                for target_layer in layers:
                    self.DC_amp[target_layer] = {}
                    for target_pop in pops:
                        self.DC_amp[target_layer][target_pop] = bg_rate*K_ext[target_layer][target_pop]*w_ext*neuron_parameters['tau_syn_E']/1000.
            else:
                self.DC_amp = DC_amp
        else:
            self.DC_amp = {'L23': {'E':0., 'I':0.},
                           'L4' : {'E':0., 'I':0.},
                           'L5' : {'E':0., 'I':0.},
                           'L6' : {'E':0., 'I':0.}}

        # scale and connect
        if establish_connections:
            # in-degrees of the full-scale model
            K_full = get_indegrees(num_neurons)

            if scale_type == 'w_and_ext':
                self.w, self.w_ext, self.DC_amp = adjust_w_and_ext_to_K(K_full, K_scaling, self.w, self.DC_amp)
                if sim.rank() == 0:
                    print 'DC_amp: ', self.DC_amp

            if scale_type == 'g_and_ext':
                self.w, self.DC_amp = adjust_g_and_ext_to_K(K_full, K_scaling, self.w, self.DC_amp)
                self.w_ext = w_ext
                if sim.rank() == 0:
                    print 'DC_amp: ', self.DC_amp

            if scale_type == 'w_and_g':
                self.w = adjust_w_and_g_to_K(K_full, K_scaling, self.w, equal_factor = w_and_g_equal_factor)
                self.w_ext = w_ext
                if numpy.any(numpy.isnan(self.w)):
                    if sim.rank() == 0:
                        print 'No solution for this scaling factor'
                    sys.exit(-1)

            if scale_type == 'w_mean_var_ext':
                self.w, K_ext_e, K_ext_i, self.DC_amp = adjust_w_mean_var_ext(K_full, self.w)
                self.w_ext = w_ext
            
            self.K_ext = { 'L23': {'E': numpy.nan, 'I': numpy.nan},
                           'L4' : {'E': numpy.nan, 'I': numpy.nan},
                           'L5' : {'E': numpy.nan, 'I': numpy.nan},
                           'L6' : {'E': numpy.nan, 'I': numpy.nan}}
            if scale_ext:
                for layer in layers:
                    for pop in pops:
                        self.K_ext[layer][pop] = K_ext[layer][pop]*K_scaling
            else:
                for layer in layers:
                    for pop in pops:
                        self.K_ext[layer][pop] = K_ext[layer][pop]

            if sim.rank() == 0:
                print "Network configuration: "
                print 'scale_type: ', scale_type
                print 'N_scaling: ', N_scaling
                print 'K_scaling: ', K_scaling
                print self.w
                if scale_type == 'w_mean_var_ext':
                    print 'in-degrees for excitatory Poisson drive: ', K_ext_e
                    print 'in-degrees for inhibitory Poisson drive: ', K_ext_i
                    print 'DC drive: ', self.DC_amp
                print 'internal in-degrees: ', K_full
                print 'external in-degrees: ', self.K_ext

            if lognormal_weights:
                w_sd = [[w_rel]*n_layers*n_pops_per_layer]*n_layers*n_pops_per_layer
            else:
                w_sd = abs(w_rel*self.w)
            possible_targets = ['inhibitory', 'excitatory']

            for target_layer in layers:
                for target_pop in pops:
                    target_index = structure[target_layer][target_pop]
                    this_pop = self.pops[target_layer][target_pop]
                    # external inputs
                    if 'DC' in input_types or scale_type in ['w_and_ext', 'g_and_ext', 'w_mean_var_ext']:
                        this_pop.set('i_offset', self.DC_amp[target_layer][target_pop])

                    if 'poisson' in input_types:
                        if nest_native_poisson:
                        # create only a single Poisson generator for each population, since the native NEST implementation sends independent spike trains to all targets
                            if scale_type == 'w_mean_var_ext':
                                sim.nest.sli_run("/poisson_generator Create /poisson_generator_e Set poisson_generator_e << /rate " + str(bg_rate*K_ext_e[target_layer][target_pop]) + " >> SetStatus")
                                sim.nest.sli_run("poisson_generator_e " + str(list(this_pop.all_cells)).replace(',', '') + " [" + str(1000*self.w_ext) + "] [" + str(d_mean['E']) + "] DivergentConnect" )
                                sim.nest.sli_run("/poisson_generator Create /poisson_generator_i Set poisson_generator_i << /rate " + str(bg_rate*K_ext_i[target_layer][target_pop]) + " >> SetStatus")
                                sim.nest.sli_run("poisson_generator_i " + str(list(this_pop.all_cells)).replace(',', '') + " [" + str(-1000*self.w_ext) + "] [" + str(d_mean['E']) + "] DivergentConnect")
                            else:
                                sim.nest.sli_run("/poisson_generator Create /poisson_generator_e Set poisson_generator_e << /rate " + str(bg_rate*K_ext[target_layer][target_pop]) + " >> SetStatus")
                                sim.nest.sli_run("poisson_generator_e " + str(list(this_pop.all_cells)).replace(',', '') + " [" + str(1000*self.w_ext) + "] [" + str(d_mean['E']) + "] DivergentConnect")

                        else:
                            # the output of a PyNN Poisson generator goes through a parrot neuron and one generator is needed for each target neuron
                            if scale_type == 'w_mean_var_ext':
                                rate_e = bg_rate*K_ext_e[target_layer][target_pop]
                                rate_i = bg_rate*K_ext_i[target_layer][target_pop]
                                if pool_poisson:
                                    poisson_pool_e = sim.Population(n_poisson_generators, sim.SpikeSourcePoisson, {'rate': poisson_rate})
                                    n_poisson_per_neuron_e = int(round(rate_e/poisson_rate))
                                    conn_e = sim.FixedNumberPreConnector(n_poisson_per_neuron_e, weights = self.w_ext)
                                    poisson_pool_i = sim.Population(this_pop.size, sim.SpikeSourcePoisson, {'rate': bg_rate*K_ext_i[target_layer][target_pop]})
                                    n_poisson_per_neuron_i = int(round(rate_i/poisson_rate))
                                    conn_i = sim.FixedNumberPreConnector(n_poisson_per_neuron_i, weights = -self.w_ext)
                                    sim.Projection(poisson_pool_e, this_pop, conn_e, target = 'excitatory')
                                    sim.Projection(poisson_pool_i, this_pop, conn_i, target = 'inhibitory')
                                else:
                                    poisson_generator_e = sim.Population(this_pop.size, sim.SpikeSourcePoisson, {'rate': rate_e})
                                    poisson_generator_i = sim.Population(this_pop.size, sim.SpikeSourcePoisson, {'rate': rate_i})
                                    conn_e = sim.OneToOneConnector(weights = self.w_ext)
                                    conn_i = sim.OneToOneConnector(weights = -self.w_ext)
                                    sim.Projection(poisson_generator_e, this_pop, conn_e, target = 'excitatory')
                                    sim.Projection(poisson_generator_i, this_pop, conn_i, target = 'inhibitory')
                            else: 
                                if pool_poisson:
                                    print 'creating pool of Poisson generators'
                                    poisson_pool = sim.Population(n_poisson_generators, sim.SpikeSourcePoisson, {'rate': poisson_rate})
                                    rate = bg_rate*self.K_ext[target_layer][target_pop]
                                    n_poisson_per_neuron = int(round(rate/poisson_rate))
                                    conn = sim.FixedNumberPreConnector(n_poisson_per_neuron, weights = self.w_ext)
                                    sim.Projection(poisson_pool, this_pop, conn, target = 'excitatory')
                                else:
                                    poisson_generator = sim.Population(this_pop.size, sim.SpikeSourcePoisson, {'rate': bg_rate*self.K_ext[target_layer][target_pop]})
                                    conn = sim.OneToOneConnector(weights = self.w_ext)
                                    sim.Projection(poisson_generator, this_pop, conn, target = 'excitatory')

                    if thalamic_input:
                    # thalamic inputs
                        if sim.rank() == 0:
                            print 'creating thalamic connections to ' + target_layer + target_pop
                        C_thal = thal_params['C'][target_layer][target_pop]
                        K_thal = round(numpy.log(1-C_thal)/numpy.log((num_neurons[target_layer][target_pop]*thal_params['n_thal']-1.)/(num_neurons[target_layer][target_pop]*thal_params['n_thal'])))/num_neurons[target_layer][target_pop]

                        if connect_type == 'multinomial':
                            if distribute_delays:
                                MultinomialConnect(sim, self.thalamic_population, this_pop, K_thal, w_mean, w_rel*w_mean, d_mean['E'], d_sd['E'])
                            else:
                                MultinomialConnect(sim, self.thalamic_population, this_pop, K_thal, w_mean, w_rel*w_mean, d_mean['E'], 0)

                    for source_layer in layers:
                        for source_pop in pops:
                            source_index = structure[source_layer][source_pop]
                            if sim.rank() == 0:
                                print 'creating connections from ' + source_layer + source_pop + ' to ' + target_layer + target_pop

                            if connect_type == 'multinomial':
                                if distribute_delays:
                                    MultinomialConnect(sim, self.pops[source_layer][source_pop], self.pops[target_layer][target_pop], K_full[target_index][source_index]*K_scaling, self.w[target_index][source_index], w_sd[target_index][source_index], d_mean[source_pop], d_sd[source_pop])
                                else:
                                    MultinomialConnect(sim, self.pops[source_layer][source_pop], self.pops[target_layer][target_pop], K_full[target_index][source_index]*K_scaling, self.w[target_index][source_index], w_sd[target_index][source_index], d_mean[source_pop], 0)

    def get_spikes(self):
        spikes = {}
        for layer in layers:
            spikes[layer] = {}
            for pop in pops:
                spikes[layer][pop] = self.pops[layer][pop].getSpikes(gather=True)
        return spikes


    def get_rates(self, spikes, t):
        rates = {}
        for layer in layers:
            rates[layer] = {}
            for pop in pops:
                if record_fraction:
                    rates[layer][pop] = 1000*len(spikes[layer][pop])/(t*frac_record_spikes*self.pops[layer][pop].size)
                else:
                    rates[layer][pop] = 1000*len(spikes[layer][pop])/(t*n_record)
        return rates


def create_weight_matrix():
    w = numpy.zeros([n_layers*n_pops_per_layer, n_layers*n_pops_per_layer])
    for target_layer in layers:
        for target_pop in pops:
            target_index = structure[target_layer][target_pop]
            for source_layer in layers:
                for source_pop in pops:
                    source_index = structure[source_layer][source_pop]
                    if source_pop == 'E':
                        if source_layer == 'L4' and target_layer == 'L23' and target_pop == 'E':
                            w[target_index][source_index] = w_234
                        else:
                            w[target_index][source_index] = w_mean
                    else:
                        if source_layer == 'L4' and target_layer == 'L4' and target_pop == 'E':
                            w[target_index][source_index] = w_4e4i
                        else:
                            w[target_index][source_index] = g*w_mean
    return w
