# Authors: Moritz Helias
#          Markus Diesmann
# Date:    2010-07-22

# Function to rescale and convert a ps file to an eps file

import fileinput
import sys
import re


def get_bounding_box(filename, bbox = "PageBoundingBox"):
    """Read the boundingbox from postscript file"""
    
    # regular expression to look for
    pattern = re.compile( "%%" + bbox + ":" )

    dim = None

    f = fileinput.input(filename, inplace=False)
    
    for line in f:
        if fileinput.isfirstline():
            sys.stderr.write('Started processing %s\n' % fileinput.filename())
      
        if pattern.match( line ):
            print "found bounding box in ", fileinput.filename(), "in line", fileinput.filelineno(), line

            # extract dimensions
            sp = line.split(' ')
            dim = map(float, sp[1:])

    f.close()
    
    return dim



def replace_bounding_box(filename, new_dimension):
    """Read the page boundingbox from postscript file"""
    
    # regular expression to look for
    pattern = re.compile( "%%BoundingBox:" )
    
    # convert list of floats to appended strings
    bbox_str = reduce ( lambda a,b : a + ' ' + b, map (str, new_dimension) )
    
    for line in fileinput.input(filename, inplace=True):
        if fileinput.isfirstline():
            sys.stderr.write('Started processing %s\n' % fileinput.filename())
      
        if pattern.match( line ):
            # found bounding box in ", fileinput.filename(), "in line", fileinput.filelineno(), line
            # now replacing with new dimensions, bbox_str
                      
            line = "%%BoundingBox: " + bbox_str
        print line.rstrip()


# this is needed because of
# usr/lib/pymodules/python2.6/matplotlib/backends/backend_ps.py, line 189, set_color
def replace_gray_with_rgb(filename):
    """Read the page boundingbox from postscript file"""
    
    # match begins with "/G{"
    # then an arbitrary number of characters not being "}" follows
    # then "}!"
    pattern = re.compile( "/G\{[^\}]*\}!" )


    for line in fileinput.input(filename, inplace=True):
        if fileinput.isfirstline():
            sys.stderr.write('Started processing %s\n' % fileinput.filename())

        sub_result = pattern.sub('/G{ dup dup rG}!', line)
        if sub_result != line:
            sys.stderr.write('found in %s\n' % line)
            line = sub_result
            sys.stderr.write(line)
            
        print line.rstrip()

def replace_setgray_with_setrgbcolor(filename):
    """Replace the occurrences of setgray with an equivalent construct using setrgb. This is needed by PloS, because they want RGB images."""
    
    pattern = re.compile( "setgray" )

    for line in fileinput.input(filename, inplace=True):
        if fileinput.isfirstline():
            sys.stderr.write('Started processing %s\n' % fileinput.filename())

        sub_result = pattern.sub('dup dup setrgbcolor', line)
        if sub_result != line:
            sys.stderr.write('found setgray in: %s\n' % line)
            line = sub_result
            sys.stderr.write('will replace by: %s' % line)
            
        print line.rstrip()
        

# usr/lib/pymodules/python2.6/matplotlib/backends/backend_ps.py, line 97, _num_to_str rounds to 3 digits
def rescale_ps_to_eps(workdir, filename, scale):
  """Rescale a ps figure by scale and create an eps."""
  
  import os
  
  # extract the PageBoundingBox: x0 y0 x1 y1 and specify these dimensions to pstops when rescaling
  # because otherwise the clippath will not be correct
  
  bbox_dim = get_bounding_box(workdir + filename, "BoundingBox")
  page_dim = bbox_dim
  print "original boundingbox is", bbox_dim

  scld_name = workdir + 'scld_' + filename
  
  # rescale the postscript file
  cmd = 'pstops -w' + str(page_dim[2])\
        + ' -h' + str(page_dim[3])\
        + ' 1:0@' + str(scale)\
        + ' ' + workdir + filename\
        + ' ' + scld_name
  
  print 'now executing ' + cmd
  os.system(cmd)


  # convert ps to eps
  cmd = 'ps2eps -f ' + scld_name
  print 'now executing ' + cmd
  os.system(cmd)

  eps_name = 'scld_' + filename.split('.')[0] + '.eps'
  print eps_name
  
  new_bbox_dim = map(lambda x: scale*x, bbox_dim)

  print "repairing bounding box to ", new_bbox_dim
  # replace the bounding box 
  replace_bounding_box(workdir + eps_name, new_bbox_dim)


# remove EOF
def merge_inkscape_eps_into_matplotlib_eps(mpl_fig_name, ink_fig_name, merged_name, position, scale):
    """Merge an eps drawing generated from inkscape into
    a matplotlib generated eps figure. (x, y) tuple position and (sx, sy) tuple
    scale specify the scaling of the graphics to merge in."""

    # open the final output file
    f_out = file(merged_name, 'wt')

    # look for showpage and remove it
    # do not remove /showpage redefinitions
    pat_showpage = re.compile('showpage$')

    # copy the figure ps code to the outfile
    for line in fileinput.input(mpl_fig_name):
        if fileinput.isfirstline():
            sys.stderr.write('Started processing %s\n' % fileinput.filename())
        
        # remove showpage
        line = pat_showpage.sub('', line)
        f_out.writelines( line )


    # look for the beginning of the prolog
    pat_header = re.compile('%%BeginProlog')

    # look for PageSetup to be removed from appended eps
    pat_beg_page_stp = re.compile('%%BeginPageSetup')

    # look for the beginning of the actual drawing
    pat_end_page_stp = re.compile('%%EndPageSetup')

    # append the drawing at the end of the matplotlib generated file
    skip = True
    for line in fileinput.input(ink_fig_name):
        if fileinput.isfirstline():    
            sys.stderr.write('Started processing %s\n' % fileinput.filename())

        # strip off the header of the drawing
        # look for %%BeginProlog and copy everything that follows
        if pat_header.match( line ):
            sys.stderr.write('found beginning of prolog, will merge what follows\n')
            f_out.writelines('%% PS_manip: begin merging file %s\n' % fileinput.filename())
            skip = False # start merging

        # remove everything
        # between BeginPageSetup and EndPageSetup
        if pat_beg_page_stp.match( line ):
            sys.stderr.write('found beginning of PageSetup, will skip what follows\n')
            skip = True # skip the page setup code

        if pat_end_page_stp.match( line ):
            sys.stderr.write('found end of page setup\n')
                
            str_trans = '%f %f translate\n' % position
            str_scale = '%f %f scale\n' % scale

            sys.stderr.write('now inserting translation and scaling\n')
            sys.stderr.write(str_trans)
            sys.stderr.write(str_scale)
            f_out.writelines('%% PS_manip: begin insert translation and scaling\n')          
            f_out.writelines([str_trans, str_scale])
            line = '%% PS_manip: end insert translation and scaling\n'           
            skip = False # start merging again
        
            
        # insert specified translation and scaling after %%EndPageSetup
        if not skip:

            f_out.writelines( line )


    f_out.writelines('%% PS_manip: end merging file %s\n' % fileinput.filename())
    f_out.close()
